import re
import time

from azureml.core import Workspace
from azureml.core.compute import AmlCompute
from azureml.core.compute import ComputeTarget

from .exceptions import AzureException
from a2ml.api.utils.decorators import error_handler, authenticated
from .credentials import Credentials


class AzureProject(object):

    def __init__(self, ctx):
        super(AzureProject, self).__init__()
        self.ctx = ctx
        self.credentials = Credentials(self.ctx).load()

    @error_handler
    @authenticated
    def list(self):
        workspaces = Workspace.list(
            self.credentials.subscription_id,
            auth=self.credentials.get_serviceprincipal_auth())
        for project in iter(workspaces.keys()):
            self.ctx.log(project)
        return {'projects': workspaces.keys()}

    @error_handler
    @authenticated    
    def create(self, name):
        name = self._get_name(name)
        region = self.ctx.config.get('cluster/region', 'eastus2')
        resource_group = self._get_resource_group(name)
        self.ctx.log('Creating %s' % name)

        self.ws = Workspace.create(
            name=name,
            subscription_id=self.credentials.subscription_id,
            resource_group=resource_group,
            create_resource_group=True,
            location=region,
            auth=self.credentials.get_serviceprincipal_auth())
        self._select(name)
        self.ctx.log('%s created' % name)
        return {'created': name}

    @error_handler
    @authenticated    
    def delete(self, name):
        name = self._get_name(name)
        ws = Workspace.get(
            name, 
            subscription_id=self.credentials.subscription_id, 
            auth=self.credentials.get_serviceprincipal_auth())
        self.ctx.log('Deleting %s' % name)
        ws.delete(delete_dependent_resources=True, no_wait=False)
        self._select(None)
        self.ctx.log('%s deleted' % name)
        return {'deleted': name}

    @error_handler
    @authenticated    
    def select(self, name = None):
        self._select(name)
        self.ctx.log('Selected project %s' % name)
        return {'selected': name}

    @error_handler
    @authenticated    
    def get_cluster_config(self, name, local_config = True, ws = None):
        result = {
            'name': self._fix_cluster_name(self.ctx.config.get('cluster/name', 'cpucluster'))
        }
        if local_config:
            result.update({
                'min_nodes': int(self.ctx.config.get('cluster/min_nodes',1)),
                'max_nodes': int(self.ctx.config.get('cluster/max_nodes',4)),
                'vm_size': self.ctx.config.get('cluster/type','STANDARD_D2_V2')
            })
            if self.ctx.config.get('cluster/idle_seconds_before_scaledown'):
                result['idle_seconds_before_scaledown'] = self.ctx.config.get('cluster/idle_seconds_before_scaledown')
        else:
            if ws is None:
                ws = self._get_ws(name=name)

            if result['name'] in ws.compute_targets:
                compute_target = ws.compute_targets[result['name']]
                if compute_target and type(compute_target) is AmlCompute:
                    #scale_settings: {'minimum_node_count': 0, 'maximum_node_count': 4, 'idle_seconds_before_scaledown': 120}
                    ct_status = compute_target.get_status()
                    if ct_status:
                        result.update({
                            'min_nodes': ct_status.scale_settings.minimum_node_count,
                            'max_nodes': ct_status.scale_settings.maximum_node_count,
                            'vm_size': ct_status.vm_size,
                            'idle_seconds_before_scaledown': ct_status.scale_settings.idle_seconds_before_scaledown
                        })

        return result
                
    @error_handler
    @authenticated    
    def update_cluster_config(self, name, params, ws=None, allow_create=True):
        cluster_name = self._fix_cluster_name(self.ctx.config.get('cluster/name', 'cpucluster'))
        if ws is None:
            ws = self._get_ws(name=name)

        if 'type' in params:
            params['vm_size'] = params['type']

        if cluster_name in ws.compute_targets:
            compute_target = ws.compute_targets[cluster_name]
            remote_cluster = self.get_cluster_config(name=name, local_config=False, ws=ws)
            update_properties = {}
            props_to_update = ['min_nodes', 'max_nodes', 'vm_size', 'idle_seconds_before_scaledown']
            for prop in props_to_update:
                if params.get(prop) is not None and remote_cluster.get(prop, params.get(prop)) != params.get(prop):
                    update_properties[prop] = params.get(prop)

            if update_properties.get('vm_size'):
                self.ctx.log('Delete existing AML compute context, since cluster type has been changed to %s.'%(update_properties.get('vm_size')))
                compute_target.delete()        
            elif update_properties:
                self.ctx.log('Update compute target %s: %s' % (cluster_name, update_properties))
                compute_target.update(**update_properties)

            try:
                compute_target.wait_for_completion(show_output = True)
            except Exception as e:
                self.ctx.log_debug(str(e))                            

            if not update_properties.get('vm_size'):
                return cluster_name

        if not allow_create:
            raise AzureException("Compute target %s does not exist."%cluster_name)

        self.ctx.log('Creating new AML compute context %s...'%cluster_name)
        provisioning_config = AmlCompute.provisioning_configuration(
            vm_size=params.get('vm_size'), min_nodes=params.get('min_nodes'),
            max_nodes=params.get('max_nodes'), idle_seconds_before_scaledown=params.get('idle_seconds_before_scaledown'))
        compute_target = ComputeTarget.create(
            ws, cluster_name, provisioning_config)
        compute_target.wait_for_completion(show_output = True)

        return cluster_name

    @staticmethod    
    def _fix_cluster_name(name):
        # Name can include letters, digits and dashes.
        # It must start with a letter, end with a letter or digit,
        # and be between 2 and 16 characters in length.
        #TODO check for all conditions

        name = re.sub(r'\W+', '-', name)
        name = name.replace('_','-')[:16]
        if name[0].isdigit():
            test = list(name)
            test[0] = 'C'
            name = ''.join(test)
        if name[-1].isdigit():
            test = list(name)
            test[-1] = 'C'
            name = ''.join(test)

        return name

    def _select(self, name):
        self.ctx.config.set('name', name)
        self.ctx.config.write()

    def _get_name(self, name = None):
        if name is None:
            name =  self.ctx.config.get('name', None)
        if name is None:
            raise AzureException('Please provide project name...')
        return name

    def _get_ws(self, name = None, create_if_not_exist = False):
        name = self._get_name(name)
        nTry = 0
        while True:
            try:
                self.ws = Workspace.get(
                    name, 
                    subscription_id=self.credentials.subscription_id, 
                    auth=self.credentials.get_serviceprincipal_auth(),
                    resource_group=self._get_resource_group(name)
                )
                break
            except Exception as e:
                message = str(e)
                if ('Workspaces not found' in message or 'No workspaces found' in message) and create_if_not_exist:
                    self.create(name)
                    break
                elif 'invalid_client' in message and nTry < 20:
                    self.ctx.log('Workspace.get failed with authentication error. Retry.')
                    nTry += 1
                    time.sleep(20) 
                else:    
                    raise

        return self.ws

    def _get_resource_group(self, name):
        resource_group = self.ctx.config.get('resource_group')
        if not resource_group:
            if name == "a2mlworkspacedev":
                resource_group = "a2mldev"
            elif name == "a2mlworkspacestaging":
                resource_group = "a2mlstaging"
            elif name == "a2mlworkspaceprod":
                resource_group = "a2mlprod"
            else:
                resource_group = name+'-resources'

        return resource_group



import pytest
from a2ml.api.a2ml import A2ML
from a2ml.api.auger.a2ml import AugerA2ML
#from a2ml.api.google.a2ml import GoogleA2ML

from .utils.mock_helpers import MockHelpers


class TestFacade(object):

    @pytest.fixture
    def mock_inits(self, monkeypatch):
        MockHelpers.pass_method(AugerA2ML, "__init__", monkeypatch)
        #MockHelpers.pass_method(GoogleA2ML, "__init__", monkeypatch)

    def test_init_a2ml(self, project, ctx, monkeypatch):
        init_auger = MockHelpers.count_calls(
            AugerA2ML, "__init__", monkeypatch)
        # init_google = MockHelpers.count_calls(
        #     GoogleA2ML, "__init__", monkeypatch)
        ctx.config.set('providers', 'auger')
        a2ml = A2ML(ctx)
        assert len(a2ml.runner.providers) == 1
        assert isinstance(a2ml.runner.providers['auger'], AugerA2ML)
        assert init_auger.times == 1
        #assert init_google.times == 0
        # modify config on the fly
        ctx.config.set('providers', ['auger'])
        init_auger.reset()
        #init_google.reset()
        a2ml = A2ML(ctx)
        assert len(a2ml.runner.providers) == 1
        assert isinstance(a2ml.runner.providers['auger'], AugerA2ML)
        #assert isinstance(a2ml.runner.providers['google'], GoogleA2ML)
        assert init_auger.times == 1
        #assert init_google.times == 1

    def test_calling_operations(self, project, ctx, mock_inits, monkeypatch):
        def test_operation(operation, args):
            auger_operation = MockHelpers.called_with(
                AugerA2ML, operation, monkeypatch)
            # google_operation = MockHelpers.called_with(
            #     GoogleA2ML, operation, monkeypatch)
            #run operation for a single provider
            ctx.config.set('providers', ['auger'])
            a2ml = A2ML(ctx)
            getattr(a2ml, operation)(*args)
            assert auger_operation.times == 1
            #assert google_operation.times == 0
            for arg in range(len(args)):
                assert auger_operation.args[arg+1] == args[arg]
            #run operation for multiple providers
            if operation != 'deploy' and operation != 'predict':
                ctx.config.set('providers', ['auger'])
                auger_operation.reset()
                #google_operation.reset()
                a2ml = A2ML(ctx)
                getattr(a2ml, operation)(*args)
                assert auger_operation.times == 1
                #assert google_operation.times == 1
                for arg in range(len(args)):
                    assert auger_operation.args[arg+1] == args[arg]
                    #assert google_operation.args[arg+1] == args[arg]
        ops = {
            'import_data': [],
            'train': [],
            'evaluate': [],
            'deploy': ['some_model_id', True],
            #'predict': ['some_model_id', 'some_csv', 0.5, True]
        }
        for opname, args in ops.items():
            test_operation(opname, args)



import aioredis
import asyncio
import json
import redis

from a2ml.server.config import Config

config = Config()

class SyncSender:
    def __init__(self):
        self.connection = None

    def publish(self, request_id, message):
        if config.notificator_redis_host:
            if not self.connection:
                self._open()

            if isinstance(message, dict):
                message = json.dumps(message)

            self.connection.xadd(request_id, {'json': message})
        else:
            # Not set NOTIFICATOR_REDIS_HOST env var if worker is run as Hub worker witouh A2ML server
            # in this case worker will not try to notify about progress
            pass

    def publish_result(self, request_id, status, result):
        self.publish(
            request_id,
            {'type': 'result', 'status': status, 'result': result}
        )

    def publish_log(self, request_id, level, msg, *args, **kwargs):
        self.publish(
            request_id,
            {'type': 'log', 'level': level, 'msg': msg, 'args': args, 'kwargs': kwargs}
        )

    def _open(self):
        self.connection = redis.Redis(
            host=config.notificator_redis_host,
            port=config.notificator_redis_port
        )

    def close(self):
        if self.connection:
            self.connection.close()

    # support with
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class AsyncReceiver:
    def __init__(self, request_id, last_msg_id):
        self.connection = None
        self.request_id = request_id
        self.last_msg_id = last_msg_id

    async def _open(self):
        self.connection = await aioredis.create_redis(
            'redis://' + config.notificator_redis_host + ':' + str(config.notificator_redis_port)
        )

    async def get_message(self, timeout=5):
        if not self.connection:
            await self._open()

        res = await self.connection.xread([self.request_id], timeout=timeout, count=1, latest_ids=[self.last_msg_id])
        if len(res) > 0:
            msg = res[0]
            self.last_msg_id = msg[1]
            data = json.loads(msg[2][b'json'])
            data['_msg_id'] = self.last_msg_id.decode('utf-8')
            return json.dumps(data)
        else:
            return None

    async def close(self):
        if self.connection:
            self.connection.close()
            await self.connection.wait_closed()



import numpy as np
from sklearn.metrics import make_scorer, mean_squared_error, mean_squared_log_error, mean_absolute_error
from sklearn.metrics import SCORERS


EPSILON = 1e-10

def _error(actual: np.ndarray, predicted: np.ndarray):
    """ Simple error """
    return actual - predicted

def _percentage_error(actual: np.ndarray, predicted: np.ndarray):
    """
    Percentage error
    Note: result is NOT multiplied by 100
    """
    return _error(actual, predicted) / (actual + EPSILON)

def mae(actual: np.ndarray, predicted: np.ndarray):
    """ Mean Absolute Error """
    return np.mean(np.abs(_error(actual, predicted)))

# def normilized_mae(y_test, y_pred):
#     from sklearn.preprocessing import MinMaxScaler

#     scaler = MinMaxScaler()
#     y_test_2 = np.reshape(y_test, (-1, 1))
#     y_pred_2 = np.reshape(y_pred, (-1, 1))
#     scaler.partial_fit(y_test_2)
#     scaler.partial_fit(y_pred_2)
#     y_test1 = scaler.transform(y_test_2)
#     y_pred1 = scaler.transform(y_pred_2)
#     return mae(np.ravel(y_test1), np.ravel(y_pred1))

def mae_ex(y_test, y_pred):
    all_errors = _error(np.ravel(y_test), np.ravel(y_pred))
    over_errors = []
    under_errors = []
    for item in all_errors:
        if item < 0:
            over_errors.append(abs(item))            
        else:
            under_errors.append(abs(item))

    mae_over = 0.0
    mae_under = 0.0
    if over_errors:
        mae_over = np.mean(over_errors)
    if under_errors:
        mae_under = np.mean(under_errors)

    return mae_over, mae_under

def _naive_forecasting(actual: np.ndarray, seasonality: int = 1):
    """ Naive forecasting method which just repeats previous samples """
    return actual[:-seasonality]

def neg_rmsle_score(y_true, y_pred):
    return -np.sqrt(mean_squared_log_error(y_true, y_pred))


def neg_rmse_score(y_true, y_pred):
    return -np.sqrt(mean_squared_error(y_true, y_pred))

def neg_mase_score(actual: np.ndarray, predicted: np.ndarray, seasonality: int = 1):
    """
    Mean Absolute Scaled Error
    Baseline (benchmark) is computed with naive forecasting (shifted by @seasonality)
    """
    return -mae(actual, predicted) / mae(actual[seasonality:], _naive_forecasting(actual, seasonality))

def neg_mape_score(y_true, y_pred):
    """
    Mean Absolute Percentage Error
    Properties:
        + Easy to interpret
        + Scale independent
        - Biased, not symmetric
        - Undefined when actual[t] == 0
    Note: result is NOT multiplied by 100
    """
    #return -np.mean(np.abs(_percentage_error(y_true, y_pred)))

    #smdape Symmetric Median Absolute Percentage Error
    #return -np.mean(2.0 * np.abs(y_true - y_pred) / ((np.abs(y_true) + np.abs(y_pred)) + EPSILON))
    
    #maape - Mean Arctangent Absolute Percentage Error
    return -np.mean(np.arctan(np.abs((y_true - y_pred) / (y_true + EPSILON))))

# https://en.wikipedia.org/wiki/Mean_Directional_Accuracy
#https://gist.github.com/bshishov/5dc237f59f019b26145648e2124ca1c9
def mda_score(y_true, y_pred):
    """ Mean Directional Accuracy """
    return np.mean((np.sign(y_true[1:] - y_true[:-1]) == np.sign(y_pred[1:] - y_pred[:-1])).astype(int))

def nmae_score(y_true, y_pred):
    """ Normalized Root Mean Squared Error """
    return mean_absolute_error(y_true, y_pred) / (y_true.max() - y_true.min())

def nrmse_score(y_true, y_pred):
    """ Normalized Root Mean Squared Error """
    return np.sqrt(mean_squared_error(y_true, y_pred)) / (y_true.max() - y_true.min())

def spearman_correlation_score(y_true, y_pred):
    from scipy import stats

    return stats.spearmanr(y_true, y_pred)[0]

neg_rmsle_scorer = make_scorer(neg_rmsle_score)
neg_mase_scorer = make_scorer(neg_mase_score)
neg_mape_scorer = make_scorer(neg_mape_score)
mda_scorer = make_scorer(mda_score)
neg_rmse_scorer = make_scorer(neg_rmse_score)
nmae_scorer = make_scorer(nmae_score)
nrmse_scorer = make_scorer(nrmse_score)
spearman_correlation_scorer = make_scorer(spearman_correlation_score)

SCORERS['neg_rmsle'] = neg_rmsle_scorer
SCORERS['neg_mase'] = neg_mase_scorer
SCORERS['neg_mape'] = neg_mape_scorer
SCORERS['mda'] = mda_scorer
SCORERS['neg_rmse'] = neg_rmse_scorer
SCORERS['normalized_mean_absolute_error'] = nmae_scorer
SCORERS['normalized_root_mean_squared_error'] = nrmse_scorer
SCORERS['spearman_correlation'] = spearman_correlation_scorer



import os

from vcr import VCR

vcr = VCR(
    cassette_library_dir=os.path.join(os.getcwd(), 'tests/cassettes'),
    decode_compressed_response=True,
    match_on=['uri', 'method'],
    record_mode='once',
    ignore_hosts=('minio', 'localhost', '127.0.0.1'),
)



import sys
import click
from a2ml.api.utils.context import CONTEXT_SETTINGS
from a2ml.api.utils.context import pass_context

COMMANDS = [
  'auth',
  'new',
  'import',
  'train',
  'evaluate',
  'deploy',
  'predict',
  'project',
  'dataset',
  'experiment',
  'model',
  'server',
  'worker'
]


class A2mlCli(click.MultiCommand):
    def list_commands(self, ctx):
        return COMMANDS

    def get_command(self, ctx, name):
        try:
            if sys.version_info[0] == 2:
                name = name.encode('ascii', 'replace')
            mod = __import__('a2ml.cmdl.commands.cmd_' + name,
                             None, None, ['cli'])
        except ImportError:
            import traceback
            traceback.print_exc()
            return
        return mod.cmdl


@click.command(cls=A2mlCli, context_settings=CONTEXT_SETTINGS)
@click.version_option()
@pass_context
def cmdl(ctx):
    """A2ML command line interface."""



import re
import ruamel.yaml

from a2ml.api.utils import fsclient


class ConfigYaml(object):
    def __init__(self):
        super(ConfigYaml, self).__init__()
        self.filename = None
        self.yaml = ruamel.yaml.comments.CommentedMap()

    def load_from_file(self, filename):
        if not isinstance(filename, str) or len(filename) == 0:
            raise ValueError("please provide yaml file name")
        self.filename = filename
        with fsclient.open_file(filename, 'r') as f:
            self.yaml = ruamel.yaml.YAML(typ='rt').load(f)
        return self

    def get(self, path, default=None):
        options = self.yaml
        if (options == None):
            return default

        path = path.split('/')
        for opt in path:
            if options is not None and opt in options:
                options = options[opt]
            else:
                return default

        if options is None:
            options = default
        return options

    def set(self, path, value):
        options = self.yaml
        path = path.split('/')
        for opt in path[0:-1]:
            if not opt in options:
                options[opt] = ruamel.yaml.comments.CommentedMap()

            options = options[opt]
                    
        options[path[-1]] = value

    def remove(self, path):
        options = self.yaml
        path = path.split('/')
        for opt in path[0:-1]:
            if opt in options:
                options = options[opt]
            else:
                return

        options.pop(path[-1], None)

    def write(self, filename=None):
        filename = filename if filename else self.filename
        with fsclient.open_file(filename, 'w') as f:
            ruamel.yaml.YAML(typ='rt').dump(self.yaml, f)
